package handlers

import (
	"context"
	"database/sql"
	"errors"
	"strings"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/template_management"
	"orus.io/orus-io/rednerd/utils"
)

// ErrDuplicateTemplateName is returned if the template's name already exists.
var ErrDuplicateTemplateName = errors.New("duplicate template name")

func stringSliceFilterOut(slice []string, values ...string) []string {
	out := make([]string, 0, len(slice)-len(values))

	for _, s := range slice {
		isPresent := false

		for _, v := range values {
			if s == v {
				isPresent = true

				break
			}
		}

		if !isPresent {
			out = append(out, s)
		}
	}

	return out
}

// NewAccountTemplateUpdateHandler returns a AccountTemplateUpdateHandler.
func NewAccountTemplateUpdateHandler(db *sqlx.DB) *AccountTemplateUpdateHandler {
	columns := stringSliceFilterOut(
		models.TemplateDBDataColumns,
		models.TemplateDBInternalVersionColumn,
		models.TemplateDBCreatedAtColumn,
	)

	return &AccountTemplateUpdateHandler{
		db:      db,
		columns: columns,
	}
}

// AccountTemplateUpdateHandler handles GET /user.
type AccountTemplateUpdateHandler struct {
	db      *sqlx.DB
	columns []string
}

func (h *AccountTemplateUpdateHandler) handle(
	ctx context.Context, account, name string, template models.Template,
) (*models.Template, error) {
	log := zerolog.Ctx(ctx)

	templatedb := models.TemplateDB{Template: template}
	if err := templatedb.ToDB(); err != nil {
		if errors.Is(err, models.ErrBadBodyFormat) {
			return nil, BadRequestError(err)
		}

		return nil, err
	}

	q := squirrel.Update(models.TemplateDBTable).
		SetMap(database.ValuesMap(templatedb, h.columns...)).
		Set(
			models.TemplateDBInternalVersionColumn,
			squirrel.Expr(models.TemplateDBInternalVersionColumn+" + 1"),
		).
		Where(squirrel.Eq{
			models.TemplateDBAccountColumn: account,
			models.TemplateDBNameColumn:    name,
		}).
		Suffix("RETURNING " + strings.Join(models.TemplateDBColumns, ","))

	tx, err := database.Begin(ctx, h.db)
	if err != nil {
		return nil, err
	}

	defer tx.RollbackIfOpened(*log)

	if err := database.GetContext(ctx, tx, &templatedb, q, log); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return nil, utils.NewHTTPErrorf(404, "No such template")
		}

		if models.IsDuplicateError(err) {
			return nil, BadRequestError(ErrDuplicateTemplateName)
		}

		return nil, err
	}

	templatedb.FromDB()
	templatedb.ForceBase64()

	if err := tx.Commit(); err != nil {
		return nil, err
	}

	if err := HandleTemplateActivity(ctx, h.db, &templatedb, log); err != nil {
		return nil, err
	}

	return &templatedb.Template, nil
}

// Handle handles the incoming request.
func (h *AccountTemplateUpdateHandler) Handle(
	params template_management.AccountTemplateUpdateParams, principal *models.Principal,
) middleware.Responder {
	ctx := params.HTTPRequest.Context()

	log := zerolog.Ctx(ctx)
	if principal.IsAdmin {
		log.Error().Msg("admin cannot manage templates. Please use a non-admin user")

		return template_management.
			NewAccountTemplateAddUnauthorized()
	}

	if principal.Subject != params.Account {
		log.Error().Msg("user can manage its own templates only")

		return template_management.
			NewAccountTemplateAddUnauthorized()
	}

	if params.Template.Account == "" {
		params.Template.Account = params.Account
	}

	if params.Template.Name == "" {
		params.Template.Name = params.Name
	}

	template, err := h.handle(
		params.HTTPRequest.Context(),
		params.Account,
		params.Name,
		*params.Template)
	if err != nil {
		var httpErr utils.HTTPError
		if errors.As(err, &httpErr) {
			return HTTPErrorResponder{httpErr}
		}

		var httpError *HTTPErrorResponder
		if errors.As(err, &httpError) {
			return httpError
		}

		return InternalError(err, log, false)
	}

	return template_management.NewAccountTemplateUpdateOK().WithPayload(template)
}
